import streamlit as st
import pickle
import sys
from pathlib import Path
import re
import streamlit.components.v1 as components
import plotly.graph_objects as go
BASE_DIR = Path(__file__).parent
CSS_PATH = BASE_DIR / "assets" / "style.css"

plot_backgroud_color = "white"
plot_font = dict(color='#989898', size=12)
plot_title_font = dict(
    color="#6B7F8A",
    family="Inter, Segoe UI, sans-serif",  
    size=16,
    style="normal",
    weight=300            
)



#Rashomon Set dashboard for binary task type

def run_app_from_file(plots_file):
    with open(plots_file, "rb") as f:
        plots, descriptions = pickle.load(f)
    
    st.set_page_config(page_title="Rashomon Dashboard", layout="wide")

    with open(CSS_PATH) as f:
        css = f.read()
    
    st.markdown(f"<style>{css}</style>", unsafe_allow_html=True)

    x0, x1, y0, y1 = 0, 1, 0, 1
    r = 0.02
    path = f'M{x0+r},{y0} ' \
       f'L{x1-r},{y0} Q{x1},{y0} {x1},{y0+r} ' \
       f'L{x1},{y1-r} Q{x1},{y1} {x1-r},{y1} ' \
       f'L{x0+r},{y1} Q{x0},{y1} {x0},{y1-r} ' \
       f'L{x0},{y0+r} Q{x0},{y0} {x0+r},{y0} Z'
    
    x0v, x1v, y0v, y1v = 0, 1, -0.02, 1.05
    path_violin = f'M{x0v+r},{y0v} ' \
       f'L{x1v-r},{y0v} Q{x1v},{y0v} {x1v},{y0v+r} ' \
       f'L{x1v},{y1v-r} Q{x1},{y1v} {x1v-r},{y1v} ' \
       f'L{x0v+r},{y1v} Q{x0v},{y1v} {x0v},{y1v-r} ' \
       f'L{x0v},{y0v+r} Q{x0v},{y0v} {x0v+r},{y0v} Z'

    title_container = st.container()
    title_container.markdown('<div class="dashboard-title">Rashomon Set Analysis</div>', unsafe_allow_html=True)
    st.markdown("<br>", unsafe_allow_html=True)
    
    #TABLES ------------------------------------------------------------------------------------------------------
    with st.container():
        rashomon_set_col, basics_col = st.columns(2)
        
        with rashomon_set_col:
            
            rashomon_table = plots["generate_rashomon_set_table"]
            rashomon_table.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    ),
                    ],
                xaxis=dict(showgrid=False, zeroline=False),
                yaxis=dict(showgrid=False, zeroline=False),
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=450,
                height=350
            )
            rashomon_table.update_layout(

                margin=dict(t=10, l=10, r=10, b=0)
            )
            st.markdown('<div class = "section_title"> Models present in the Rashomon Set</div>', unsafe_allow_html=True)
            with st.container(key = "table_container", height="stretch"):
                st.plotly_chart(rashomon_table, width='stretch')
                st.markdown(f'<div class="plots_descr rashomon-table-descr">{descriptions["generate_rashomon_set_table"]}</div>', unsafe_allow_html=True)
        
        
        with basics_col:
            st.markdown(
                        '<div class="section_title">Basic properties of the Rashomon Set</div>',
                        unsafe_allow_html=True
                    )
            with st.container(key='basics', height="stretch"):
                
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                            <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px; ">Base model : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["base_model_return"]}</span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                             <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px;">Base metric : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["base_metric_return"]}</span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                            <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px;">Base model's score : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["base_model_score_return"]:.3f}</span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                                
                gauge_size = plots['set_size_indicator']
                gauge_size.update_layout(
                    height = 280,
                    paper_bgcolor=plot_backgroud_color,
                    plot_bgcolor=plot_backgroud_color,
                    margin=dict(t=50, l=80, r=80, b=10),
        
                )
                
                gauge_descr = descriptions['set_size_indicator']

            
                st.plotly_chart(gauge_size)
                st.markdown(
                        f"""
                        <div class = "plots_descr">
                         {gauge_descr}
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                    
            

    
    with st.container():
        st.markdown(
            '<div class="section_title">Rashomon Ratio and Pattern Ratio metrics</div>',
            unsafe_allow_html=True
        )
        rashomon_ratio_col,pattern_ratio_col = st.columns(2)

        with pattern_ratio_col:              
            # # PATTERN RATIO INDICATOR
            pattern_indicator, pattern_indicator_descr = plots['pattern_ratio_indicator'], descriptions['pattern_ratio_indicator']
            pattern_indicator.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=75,
                font=plot_font,
            )
            pattern_indicator.update_layout(
                title={
                    'text': "Pattern Rashomon Ratio",
                    'y':0.97,
                    'x':0.05,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=10, r=10, b=25)
            )
    
            with st.container(key = "pattern_white_container", height="stretch"):
                    st.plotly_chart(pattern_indicator, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{pattern_indicator_descr}</div>',
                        unsafe_allow_html=True
                    )
    

        with rashomon_ratio_col:
        # RASHOMON RATIO 
            ratio_indicatior, ratio_indicator_descr = plots['rashomon_ratio_indicator'], descriptions['rashomon_ratio_indicator']
            ratio_indicatior.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                font=plot_font,
            )
            ratio_indicatior.update_layout(
                title={
                    'text': "Rashomon Ratio",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                width=75,
                height=75,
                margin=dict(t=30, l=10, r=10, b=25)
            )

            with st.container(key = "ratio_white_container", height="stretch"):
                    st.plotly_chart(ratio_indicatior, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{ratio_indicator_descr}</div>',
                        unsafe_allow_html=True
                    )
           
        
        

# RATIOS VS EPSILONS
    with st.container():
        col1, col2 = st.columns(2)
        with col1:
             # # RASHOMON RATIO VS EPSILON
            ratio_vs_eps_plot, ratio_vs_eps_descr = plots['rashomon_ratio_vs_epsilon'], descriptions['rashomon_ratio_vs_epsilon']
            ratio_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor= plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=300,
                font=plot_font,
                xaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color='#989898')),
                yaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color='#989898')),
                legend=dict(font=dict(color='#989898', size=10)),
                title={
                    'text': "Rashomon Ratio vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=50, l=10, r=10, b=0)
            )
            


            with st.container(key = "ratio_epsilon_white_container", height="stretch"):
                    st.plotly_chart(ratio_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{ratio_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )
        with col2:
            # PATTERN VS EPSILON
            pattern_vs_eps_plot, pattern_vs_eps_descr = plots['pattern_rashomon_ratio_vs_epsilon'], descriptions['pattern_rashomon_ratio_vs_epsilon']
            pattern_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=300,
                font=plot_font,
                xaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color='#989898')),
                yaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color='#989898')),
                legend=dict(font=dict(color='#989898', size=10)),
                title={
                    'text': "Pattern Ratio vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=50, l=10, r=10, b=0)
            )

            with st.container(key = "pattern_epsilon_white_container", height="stretch"):
                    st.plotly_chart(pattern_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{pattern_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )


    #AMBIGUITY AND DISCREPANCY -------------------------------------------------------------------------------------------------------
    #with st.container(key="amb_discr"):

    with st.container():
        st.markdown(
            '<div class="section_title">Ambiguity and Discrepancy metrics</div>',
            unsafe_allow_html=True
        )
        descr= """Ambiguity can be interpreted as the proportion of observations that would receive a different prediction if we were to choose another model from the Rashomon Set, while
        Discrepancy is the maximum proportion of individuals whose predictions would change if we choose a different competing model from the Rashomon Set."""
        st.markdown(
            f'<div class="section_description">{descr}</div>',
            unsafe_allow_html=True
        )
        st.markdown(
            """
            <style>
            [data-baseweb="checkbox"] [data-testid="stWidgetLabel"] p {
                /* Styles for the label text */
                color: #8B9DA8;
                font-family: 'Inter', sans-serif; 
                font-size: 16px; 
                font-weight: 300 !important;
                letter-spacing: 0.1px;
                line-height: 1.5; 

            }
            </style>
            """,
            unsafe_allow_html=True

        )

        use_probability = st.checkbox("Show probabilistic version", value=False)
        
        if use_probability:
            st.markdown( f'<div class="section_description">While calculating probabilistic ambiguity and discrepancy, the difference between risk probabilities is considered meaningful if it is greater then or equal to delta.</div>',
            unsafe_allow_html=True)

        lolipop_col, amb_col, discr_col = st.columns([1.1,1.5,1.5])
        with st.container():
            with lolipop_col:
                
                if use_probability:
                    amb_discr_plot, amb_discr_descr = plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"]
                else:
                    amb_discr_plot, amb_discr_descr = plots['lolipop_ambiguity_discrepancy'], descriptions['lolipop_ambiguity_discrepancy']

                amb_discr_plot.update_layout(
                    paper_bgcolor= plot_backgroud_color,
                    plot_bgcolor=plot_backgroud_color,
                    height=250,
                    font=plot_font,
                    xaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color="#989898")),
                    yaxis=dict(title_font=dict(color='#989898'), tickfont=dict(color='#989898')),
                    legend=dict(font=dict(color='#989898', size=10)),
                    title={
                    'text': "Ambiguity and Discrepancy values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                    },
                )
                with st.container(key = "lolipop_container"):
                    st.plotly_chart(amb_discr_plot)
                    st.markdown(
                        f'<div class ="plots_descr">{amb_discr_descr}</div>',
                        unsafe_allow_html=True
                    )

            with amb_col:
                if use_probability:
                    amb_vs_eps_plot, amb_vs_eps_descr = plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"]
                else:
                    amb_vs_eps_plot, amb_vs_eps_descr = plots['ambiguity_vs_epsilon'], descriptions['ambiguity_vs_epsilon']
                amb_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=250,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Ambiguity vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=5, r=5, b=5)
                )

                with st.container(key="ambig_epsilon_container", height="stretch"):
                    st.plotly_chart(amb_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{amb_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )

            with discr_col:
                if use_probability:
                    discr_vs_eps_plot, discr_vs_eps_descr = plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"]
                else:
                    discr_vs_eps_plot, discr_vs_eps_descr = plots['discrepancy_vs_epsilon'], descriptions['discrepancy_vs_epsilon']
                discr_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=250,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Discrepancy vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=5, r=5, b=5)
                )

                with st.container(key = "discr_epsilon_container", height="stretch"):
                    st.plotly_chart(discr_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{discr_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )

    #RASHOMON CAPACITY AND PER SAMPLE -------------------------------------------------------------------------------------------------------
    with st.container():
        left, right = st.columns([45,55], vertical_alignment="top")
        with right:
            st.markdown('<div class = "section_title">Single sample analysis</div>', unsafe_allow_html=True)
            descr= f"""
            This section presents a detailed analysis of a randomly selected sample from the dataset.
            <br>
            If you wish to explore a specific sample, try experimenting with the methods (rashomon_capacity_for_sample and proba_probabilities_for_sample) available in the Visualizer class instance.
            </br>
            """
            st.markdown(
                    f'<div class ="section_description">{descr}</div>',
                    unsafe_allow_html=True
                )
            
        with left:
            st.markdown(
                '<div class="section_title">Rashomon Capacity metric</div>',
                unsafe_allow_html=True
            )
            no_classes = descriptions["number_of_classes_return"]
            descr= f"""Rashomon Capacity is a predictive multiplicity metric that measures the diversity of model predictions for each sample.
            <br> 
            <strong>Values are between [1, {no_classes}]</strong>, where 1 indicates do predictive multiplicty present, and {no_classes} represents the highest possible diversity among model predictions.
            </br>
            """
            st.markdown(
                f'<div class="section_description">{descr}</div>',
                unsafe_allow_html=True
            )
    with st.container():
        
        rashomon_capacity_col, per_sample_col = st.columns([45,55], vertical_alignment="top")

        with rashomon_capacity_col:
            with st.container():
                capacity_hist, capacity_hist_descr = plots['rashomon_capacity_distribution'], descriptions['rashomon_capacity_distribution']
                capacity_hist.update_layout(
                    shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=230,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Rashomon Capacity distribution",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=40, l=5, r=10, b=5)
                )

                capacity_boxplot, capacity_boxplot_descr = plots['rashomon_capacity_distribution_by_class'], descriptions['rashomon_capacity_distribution_by_class']
                capacity_boxplot.update_layout(
                    shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                    ],
                    paper_bgcolor = plot_backgroud_color,
                    plot_bgcolor=plot_backgroud_color,
                    height=230,
                    font=plot_font,
                    xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    legend=dict(font=dict(color="#989898", size=10)),
                    title={
                        'text': "Rashomon Capacity distribution by class",
                        'y':0.97,
                        'x':0.02,
                        'xanchor': 'left',
                        'yanchor': 'top',
                        'font': plot_title_font
                    },
                    margin=dict(t=40, l=5, r=10, b=5)
                    )

                #histogram container for plot and description
               
                with st.container(key='capacity-hist'):
                    st.plotly_chart(capacity_hist, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_hist_descr}</div>',
                        unsafe_allow_html=True
                    )

                #boxplot container for plot and description
                with st.container(key='capacity-boxplot', height="stretch"):
                    st.plotly_chart(capacity_boxplot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_boxplot_descr}</div>',
                        unsafe_allow_html=True
                    )
                    

        with per_sample_col:
                
            proba_probabilities_plot, proba_probabilities_descr = plots['proba_probabilities_for_sample'], descriptions['proba_probabilities_for_sample']

            proba_probabilities_plot.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=400,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=10)),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=10)),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Predicted probabilities for sample across all models in the Rashomon Set",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=30, l=5, r=5, b=5)
            )

            capacity_per_sample, capacity_per_sample_descr = plots['rashomon_capacity_for_sample'], descriptions['rashomon_capacity_for_sample']

            capacity_per_sample.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=100,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Rashomon Capacity value for sample",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=60, l=5, r=5, b=10)
            )

            
            #proba probabilities container for plot and description
            with st.container(key='per_sample', height="stretch"):
                st.plotly_chart(proba_probabilities_plot, width='stretch')
                st.markdown(
                    f'<div class="plots_descr">{proba_probabilities_descr}</div>',
                    unsafe_allow_html=True
                )

            #rashomon capacity for sample container for plot and description
                with st.container(key="capacity_per_sample_padding"):
                    st.plotly_chart(capacity_per_sample, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_per_sample_descr}</div>',
                        unsafe_allow_html=True
                    )


    #PERCENT AGREEMENT AND FEATURE IMPORTANCE -------------------------------------------------------------------------------------------------------

    with st.container():
        
        percent_agreement_bar_plot, percent_agreement_bar_descr = plots['percent_agreement_barplot'], descriptions['percent_agreement_barplot']
        percent_agreement_bar_plot.update_layout(
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=350,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Percent Agreement Rates distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=10, b=5)
            )
        
        cohens_kappa_barplot, cohens_kappa_barplot_descr = plots['cohens_kappa_diverging_barplot'], descriptions['cohens_kappa_diverging_barplot']
        cohens_kappa_barplot.update_layout(
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=350,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Cohen's Kappa distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=10, b=5)
            )

        
        agreement_rates_violin_plot, agreement_rates_violin_descr = plots['agreement_rates_density'], descriptions['agreement_rates_density']
        agreement_rates_violin_plot.update_yaxes(range=[-0.005,1.005], tickvals=[0, 0.2, 0.4, 0.6, 0.8, 1], showgrid=True)
        agreement_rates_violin_plot.update_layout(
            shapes=[
            dict(
                type="path",
                path=path_violin,
                xref="paper", yref="paper",
                line=dict(color="#6C929E", width=1),
                fillcolor="rgba(0,0,0,0)"
            )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=390,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898"), showgrid=True),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Agreement Rates distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=50, l=5, r=10, b=5)
            )
        
        fi_table, fi_table_descr = plots['feature_importance_table'], descriptions['feature_importance_table']
        fi_heatmap, fi_heatmap_descr = plots['feature_importance_heatmap'], descriptions['feature_importance_heatmap']

        fi_table.update_traces(columnwidth=[1,2,2])
        fi_table.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                ),
                ],
            xaxis=dict(showgrid=False, zeroline=False),
            yaxis=dict(showgrid=False, zeroline=False),
            paper_bgcolor= plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=139
        )
        fi_table.update_layout(
            title={
                'text': "Top 3 Feature Importance in Rashomon Set",
                'y':0.97,
                'x':0.05,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=10, r=10, b=10)
        )

        heatmap_is_empty = (len(fi_heatmap.data) == 0 or fi_heatmap.data[0].z is None or len(fi_heatmap.data[0].z) == 0)
        if not heatmap_is_empty:
            fillcolor = 'rgba(108,146,158,0.2)'
        else:
            fillcolor = 'rgba(0,0,0,0)'
                
        fi_heatmap.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                ),
                dict(
                    type="rect",
                    xref="paper", yref="paper",
                    x0=0, y0=0, x1=1, y1=1,
                    fillcolor=fillcolor,
                    line=dict(width=0),
                    layer="below"   #background below the heatmap to show borders between cells
                )
            ],
            paper_bgcolor= plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=570,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=8)),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=8)),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Feature Importance Rank",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=0, r=5, b=5),
            )

        percent_agreement_col, feature_importance_col = st.columns([45, 55], vertical_alignment='top')
        agreement_rates_col, fi_heatmap_col = st.columns([45, 55], vertical_alignment='bottom')

        with st.container():

            with percent_agreement_col:
                st.markdown('<div class = "section_title">Models agreement metrics</div>', unsafe_allow_html=True)
                descr = '''
                    This section shows how models in the Rashomon Set agree with the base model using Percent Agreement or Cohen’s Kappa metrics.
                    '''
                st.markdown(f'<div class="section_description no_padding">{descr}</div>', unsafe_allow_html=True)

                metric_option = st.radio(
                    "Select metric:",
                    ["Percent Agreement", "Cohen's Kappa"],
                    horizontal=True,
                    label_visibility="collapsed"
                )

                if metric_option=="Percent Agreement":
                    st.markdown('<div class = "section_title">Percent Agreement</div>', unsafe_allow_html=True)
                    descr = '''
                        Percent Agreement is defined as the proportion of observations for which each model from the Rashomon Set predicted the same class as the base model.
                        '''
                    st.markdown(f'<div class="section_description less_padding">{descr}</div>', unsafe_allow_html=True)
                    with st.container(key = "percent_agreement"):
                        st.plotly_chart(percent_agreement_bar_plot, width='stretch')
                        st.markdown(f'<div class="plots_descr">{percent_agreement_bar_descr}</div>', unsafe_allow_html=True)

                elif metric_option=="Cohen's Kappa":
                    st.markdown('<div class = "section_title">Cohens Kappa</div>', unsafe_allow_html=True)
                    descr = '''
                        Cohen's Kappa measures how much each model agrees with the base model beyond chance.
                        Values range from -1 (complete disagreement) to 1 (perfect agreement), with 0 meaning chance-level agreement.
                        '''
                    st.markdown(f'<div class="section_description less_padding">{descr}</div>', unsafe_allow_html=True)
                    with st.container(key = "cohens_kappa_barplot"):
                        st.plotly_chart(cohens_kappa_barplot, width='stretch')
                        st.markdown(f'<div class="plots_descr">{cohens_kappa_barplot_descr}</div>', unsafe_allow_html=True)


            with feature_importance_col:
                st.markdown('<div class = "section_title">Feature Importance</div>', unsafe_allow_html=True)
                descr = '''
                This section provides an overview of the most important features identified across the models included in the Rashomon Set, highlighting which features contribute most to the predictions of each model.
                <br>
                <em>Note that this section presents information for models with feature importnace data available. 
                If feature importance was not selected or was unavailable for certain models, no information is displayed.</em><br>
                '''
                st.markdown(f'<div class="section_description additional_padding">{descr}</div>', unsafe_allow_html=True)
                with st.container(key='fi_table'):
                    st.plotly_chart(fi_table, width='stretch')
                    fi_table_description_update = fi_table_descr + " <br>If the base model does not provide feature importance values, a ‘–’ will be displayed to indicate that this information is unavailable."
                    st.markdown(f'<div class = "plots_descr">{fi_table_description_update}</div>', unsafe_allow_html=True)
                st.markdown(f'<div class="section_description heatmap-descr">{fi_heatmap_descr}</div>', unsafe_allow_html=True)

        with st.container():
            with agreement_rates_col:
                st.markdown('<div class ="section_title">Agreement Rates</div>', unsafe_allow_html=True)
                descr = '''
                For each observation, the Agreement Rate quantifies the proportion of models that produce the same class prediction as the base model.
                '''
                st.markdown(
                        f'<div class="section_description">{descr}</div>',
                        unsafe_allow_html=True
                    )

                #violin container for plot and description
                with st.container(key='agreement_rates_violin'):
                    st.plotly_chart(agreement_rates_violin_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{agreement_rates_violin_descr}</div>',
                        unsafe_allow_html=True
                    )
            with fi_heatmap_col:
                #feature importance heatmap
                with st.container(key = 'fi_heatmap', height="stretch"):
                    st.plotly_chart(fi_heatmap, width='stretch')

                    
    #VPRS -------------------------------------------------------------------------------------------------------
    with st.container():
        with st.container():
            st.markdown('<div class = "section_title">Viable Prediction Range</div>', unsafe_allow_html=True)

            vprs_hist_plot, vprs_hist_descr = plots['vpr_width_histogram'], descriptions['vpr_width_histogram']
            vprs_by_class_plot, vprs_by_class_descr = plots['vprs_widths_plot'], descriptions['vprs_widths_plot']

            plot_description, vpr_definition = vprs_by_class_descr.split('.', 1)
            plot_description = plot_description + '.' 
            vpr_definition = vpr_definition.strip()

            descr= f"""
            {vpr_definition}
            """
            st.markdown(
                    f'<div class = "section_description">{descr}</div>',
                    unsafe_allow_html=True
                )

            vprs_hist_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    ),
                    ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=250,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Viable Prediction Range Widths",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=5, r=5, b=5)
                )

            vprs_by_class_plot, vprs_by_class_descr = plots['vprs_widths_plot'], descriptions['vprs_widths_plot']

            vprs_by_class_plot.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=270,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Viable Prediction Range Widths by class",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=30, l=5, r=5, b=5)
            )
            vprs_by_class_plot.update_xaxes(
                showticklabels=False,  # usuwa numery/tick labels
                showgrid=False,        # usuwa linie siatki dla osi X, jeśli chcesz
                zeroline=False         # usuwa linię 0, jeśli jest
            )

            vprs_vs_base_model_plot, vprs_vs_base_model_plot_descr = plots['vpr_vs_base_model_plot'], descriptions['vpr_vs_base_model_plot']
            vprs_vs_base_model_plot.update_yaxes(range=[-0.005,1.005], tickvals=[0, 0.2, 0.4, 0.6, 0.8, 1], showgrid=True)
            vprs_vs_base_model_plot.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path_violin,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=350,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898"), showgrid=True),
            showlegend=False,
            title={
                'text': "Viable Prediction Range Widths by class",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=5, b=5)
            )

            hist_col, by_class_col = st.columns([1, 1])
            with hist_col:
                #vpr widths hist container for plot and description
                with st.container(key='vprs_hist', height="stretch"):
                    st.plotly_chart(vprs_hist_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{vprs_hist_descr}</div>',
                        unsafe_allow_html=True
                    )
            
            with by_class_col:
                #vpr widths by class container for plot and description
                with st.container(key='vprs_by_class', height="stretch"):
                    st.plotly_chart(vprs_by_class_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{plot_description}</div>',
                        unsafe_allow_html=True
                    )
            
            with st.container(key='vprs_vs_base_model', height="stretch"):
                chart_col, descr_col = st.columns([3, 1])
                with chart_col:
                    st.plotly_chart(vprs_vs_base_model_plot, width='stretch')
                with descr_col:
                    st.markdown('<br>', unsafe_allow_html=True)
                    st.markdown(f'<div class="plots_descr">{vprs_vs_base_model_plot_descr}</div>', unsafe_allow_html=True)
                            
if __name__ == "__main__":
    plots_file = sys.argv[1]   
    run_app_from_file(plots_file)


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator
import seaborn as sns
import warnings

from .rashomon_set import *

class RashomonIntersection(RashomonSet):
    '''
        Class for calculating Rashomon set intersection for multiple Rashomon sets based on different metrics.
        Inherits from RashomonSet class.
    '''

    def __init__(self, leaderboard : pd.DataFrame, predictions : dict, proba_predictions : dict, feature_importances : dict,  metrics : list, epsilon: float,  custom_weights : list =None, weighted_sum_method : str = 'entropy'):
        '''
            Initializes RashomonSetIntersection with two Rashomon sets.

            Args:
                leaderboard : dataframe with all trained model names and their test scores (returned by Converter)
                predictions : dictionary with model + prediction vector (returned by Converter)
                proba_predictions : dictonary with model + proba predictions vector (returned by Converter)
                feature_importances : dictionary with model + sorted feature names by their feature importance (returned by Converter)
                metrics : list of metrics to be used in the intersection calculation
                epsilon : epsilon value informing how the test scores can differ in order for models to be included in rashomon set 
                custom_weights : if weighted_sum is 'custom_weights' then user must specify weights in 2-element list.
                weighted_sum_method : Specifies the weighted sum method for selecting the base model. Options are:
                    - None or 'entropy' (default): Uses entropy-based method to select the base model.
                    - 'critic': Uses critic-based method to select the base model.
                    - 'custom_weights' : Uses the user specified weights to select the base model.
                
        '''

        #metrics validation
        if not isinstance(metrics, list):
            raise TypeError(f"'metrics' must be a list, got {type(metrics).__name__} instead.")
        if len(metrics)!=2:
            raise ValueError('This class calculates intersection for 2 given metrics. Please pass on a list of metrics with the correct length.')
        if metrics[0] not in self.METRICS:
            raise ValueError(f"Metric '{metrics[0]}' is not supported. Supported metrics are: {self.METRICS}")
        if metrics[1] not in self.METRICS:
            raise ValueError(f"Metric '{metrics[1]}' is not supported. Supported metrics are: {self.METRICS}")
        if metrics[0] == metrics[1]:
            raise ValueError('Metrics should be different.')
        
        #leaderboard validation -> not empty, pd.DataFrame type, must contain columns 'model' and metrics, metrics columns cannot contain NaN values
        if not isinstance(leaderboard, pd.DataFrame):
            raise(TypeError(f'leaderboard must be a pandas DataFrame, got {type(leaderboard)} instead.'))
        if leaderboard.empty:
            raise(ValueError('leaderboard is empty, cannot create RashomonSet object.'))
        if 'model' not in leaderboard.columns:
            raise(ValueError("leaderboard DataFrame must contain 'model' column with model names."))
        if metrics[0] not in leaderboard.columns:
            raise ValueError(f"leaderboard DataFrame must contain {metrics[0]} column.")
        if metrics[1] not in leaderboard.columns:
            raise ValueError(f"leaderboard DataFrame must contain {metrics[1]} column.")
        if leaderboard[metrics[0]].isna().any():
            raise ValueError(f"leaderboard DataFrame must contain valid metric values in {metrics[0]} column, not NaN.")
        if leaderboard[metrics[1]].isna().any():
            raise ValueError(f"leaderboard DataFrame must contain valid metric values in {metrics[1]} column, not NaN.")
        
        #predictions validation -> not empty dictionary with keys being model names, and values being predictions
        if not isinstance(predictions, dict):
            raise(TypeError(f'predictions must be a dictionary, got {type(predictions).__name__} instead.'))
        #predictions dictionary cannot be empty
        if not predictions:
            raise(ValueError('predictions dictionary is empty, cannot create RashomonSetIntersection object.'))
        #all values in predictions dict must be pandas Series and cannot be empty
        for model, preds in predictions.items():
            if not isinstance(preds, pd.Series):
                raise TypeError(f"Predictions for model '{model}' must be a pandas Series, not {type(preds).__name__}.")
            if preds.empty:
                raise ValueError(f"Predictions for model '{model}' are empty.")
        #all keys in predictions dict must match model names in leaderboard
        missing_in_leaderboard = set(predictions.keys()) - set(leaderboard['model'])
        missing_in_predictions = set(leaderboard['model']) - set(predictions.keys())
        if missing_in_leaderboard:
            raise ValueError(f"The following models in predictions dict are not found in leaderboard: {missing_in_leaderboard}")
        if missing_in_predictions:
            raise ValueError(f"The following models in leaderboard are not found in predictions dict: {missing_in_predictions}")
        
        #proba predictions validation
        if not isinstance(proba_predictions, dict):
            raise(TypeError(f'proba_predictions must be a dictionary, got {type(proba_predictions).__name__} instead.'))
        #proba predictions dictionary cannot be empty
        if not proba_predictions:
            raise(ValueError('proba_predictions dictionary is empty, cannot create RashomonSetIntersection object.'))
        #all values in proba predictions dict must be pandas DataFrame and cannot be empty
        for model, preds in proba_predictions.items():
            if not isinstance(preds, pd.DataFrame):
                raise TypeError(f"Proba predictions for model '{model}' must be a pandas DataFrame, not {type(preds).__name__}.")
            if preds.empty:
                raise ValueError(f"Proba predictions for model '{model}' are empty.")
        #all keys in proba predictions dict must match model names in leaderboard
        missing_in_leaderboard_proba = set(proba_predictions.keys()) - set(leaderboard['model'])
        missing_in_proba_predictions = set(leaderboard['model']) - set(proba_predictions.keys())
        if missing_in_leaderboard_proba:
            raise ValueError(f"The following models in proba predictions dict are not found in leaderboard: {missing_in_leaderboard_proba}")
        if missing_in_proba_predictions:
            raise ValueError(f"The following models in leaderboard are not found in proba predictions dict: {missing_in_proba_predictions}")
        
        #feature importances validation
        if feature_importances is not None:
            if not isinstance(feature_importances, dict):
                raise(TypeError(f'feature_importances must be a dictionary, got {type(feature_importances).__name__} instead.'))
            #feature importances dictionary cannot be empty
            if not feature_importances:
                raise(ValueError('feature_importances dictionary is empty, cannot create RashomonSetIntersection object.'))
            #some models do not produce feature importances, so we warn the user insted of raising an error
            #for keys in feature importances dict that are not in leaderboard, we warn that these models will be ignored
            missing_in_leaderboard_fi = set(feature_importances.keys()) - set(leaderboard['model'])
            missing_in_fi = set(leaderboard['model']) - set(feature_importances.keys())
            if missing_in_leaderboard_fi:
                warnings.warn(
                    f"The following models in feature_importances dict are not found in leaderboard "
                    f"(they will be ignored): {missing_in_leaderboard_fi}")
            if missing_in_fi:
                warnings.warn(
                    f"The following models from leaderboard are missing in feature_importances dict "
                    f"(possibly because these models do not provide feature importances): {missing_in_fi}")
                
        self.metrics = metrics
        self.custom_weights = custom_weights #user paramter value saved 
        self.weighted_sum_method = weighted_sum_method #user paramter value saved 
        self.leaderboard = leaderboard
        self.predictions_dict = predictions
        self.proba_predictions_dict = proba_predictions
        self.feature_importance_dict = feature_importances

        self.base_metric = None #not defined in Rashomon Intersection
        self.worst_score = None #not defined in Rashomon Intersection
        self.best_score = None #not defined in Rashomon Intersection

        if not isinstance(epsilon, float):
            raise(TypeError('epsilon must be a float value.'))
        if epsilon<0:
            raise(ValueError('epsilon must be non-negative float value.'))
        self.epsilon = epsilon

        self.all_rashomon_sets = {metric: self.get_rashomon_set_for_metric(metric) for metric in metrics}
        self.rashomon_set = self.find_rashomon_intersection()
        if len(self.rashomon_set)<=1:
            raise ValueError(f'For given base metrics : {self.metrics} and epsilon value = {self.epsilon}, the Rashomon Intersection consists of {len(self.rashomon_set)} models. Please provide valid epsilon and base metrics.')
        
        self.rashomon_predictions, self.rashomon_proba_predictions = self.get_rashomon_predictions()
        if feature_importances is not None:
            self.rashomon_feature_importance = self.get_rashomon_feature_importances()
        else:
            self.rashomon_feature_importance = None
        self.number_of_classes = self.determine_number_of_classes()
        self.task_type = self.determine_task_type()

        #base_model based on weighted sum method, by default entropy method
        if weighted_sum_method=='entropy':
            if custom_weights is not None:
                raise ValueError(f"For 'entropy' weighted_sum_method, 'custom_weights' must be None. Received: {custom_weights}")
            self.base_model = self.find_base_model_entropy_based()
            self.weights = self.find_weights_entropy_based()
        elif weighted_sum_method =='critic':
            if custom_weights is not None:
                raise ValueError(f"For 'critic' weighted_sum_method, 'custom_weights' must be None. Received: {custom_weights}")
            self.base_model = self.find_base_model_critic_based()
            self.weights = self.find_weights_critic_method()
        elif weighted_sum_method=='custom_weights':
            if custom_weights is None:
                raise ValueError("Custom weights method requires 'custom_weights' argument, got None.")
            if not isinstance(custom_weights, list):
                raise TypeError(f"'custom_weights' must be list, got {type(custom_weights).__name__}.")
            if len(custom_weights) != 2:
                raise ValueError(f"'custom_weights' must have exactly 2 elements, got {len(custom_weights)}.")
            if not np.isclose(sum(custom_weights), 1.0, atol=1e-4):
                raise ValueError(f"The sum of 'custom_weights' must be 1.0, got {sum(custom_weights)}.")
            self.base_model = self.find_base_model(custom_weights[0], custom_weights[1])
            self.weights = pd.Series(custom_weights)
        else:
            raise ValueError(f'Unknown weighted sum method :  {weighted_sum_method}. Try custom_weights, entropy or critic')
        

    def get_rashomon_set_for_metric(self, metric: str, epsilon : float = None) -> list:
        '''
            Method for getting Rashomon Set for a given metric.

            Args:
                metric : metric for which Rashomon Set is to be calculated

            Returns:
                list : List of models that are in Rashomon Set for the given metric
        '''

        if metric not in self.METRICS:
            raise ValueError(f"Metric '{metric}' is not supported. Supported metrics are: {self.METRICS}")
        if epsilon is None:
            epsilon = self.epsilon
       
        if self.METRICS_GREATER_IS_BETTER.get(metric):
            best_metric_value = self.leaderboard[metric].max()
            rashomon_models = self.leaderboard[self.leaderboard[metric] >= best_metric_value - epsilon]
        else:
            best_metric_value = self.leaderboard[metric].min()
            rashomon_models = self.leaderboard[self.leaderboard[metric] <= best_metric_value + epsilon]

        rashomon_models_names = rashomon_models['model']

        return rashomon_models_names
    

    def find_rashomon_intersection(self, epsilon :float = None) -> list:
        '''
            Method for calculating intersection of Rashomon Sets based on the models from Rashomon Sets chosen for each test metric.

            Returns:
                list : List of model names that are in intersection of Rashomon Sets
        '''
        if epsilon is None:
            epsilon = self.epsilon

        rashomon_sets = [self.get_rashomon_set_for_metric(metric, epsilon) for metric in self.metrics]
        if not rashomon_sets:
            return []
        else:
            return list(set.intersection(*map(set, rashomon_sets)))
        

    def get_rashomon_set(self, epsilon : float = None) -> list:
        '''
            Method for finding all models that are included in the Rashomon Intersection. 

            Returns:
                list : List of model names that are in the Rashomon Intersection
        '''

        if epsilon is None:
            epsilon = self.epsilon

        rashomon_intersection = self.find_rashomon_intersection(epsilon)
        return rashomon_intersection
        
    def get_rashomon_predictions(self)-> Tuple[Dict,Dict]:
        '''
            Override method to use intersection models, not single metric models.
        '''

        rashomon_predictions = {model: self.predictions_dict[model] for model in self.rashomon_set}
        rashomon_proba_predictions = {model: self.proba_predictions_dict[model] for model in self.rashomon_set}
        return rashomon_predictions, rashomon_proba_predictions
    
    def get_rashomon_feature_importances(self):
        '''
            Override method to use intersection models, not single metric models.
        '''
        rashomon_importances = {model : self.feature_importance_dict[model] for model in self.rashomon_set}
        return rashomon_importances
        
    
    def find_base_model(self, weight1, weight2) -> str:
        '''
            Method for finding base_model based on user passed weight values. 

            Returns:
                str : Model name that maximizes the sum : weight1*metrics[0] + weight2*metrics[1].
                      If many models have the same sum value, base model is the first one in the leaderboard.
        '''

        if not np.isclose(weight1+weight2, 1, atol=1e-6) or weight1<0 or weight2<0:
            raise ValueError('Passed weights should be >= 0 and sum to 1')
        
        rashomon_leaderboard_subset = self.leaderboard[self.leaderboard['model'].isin(self.rashomon_set)].copy()
        rashomon_leaderboard_subset['weighted_sum'] = weight1 * rashomon_leaderboard_subset[self.metrics[0]] + weight2* rashomon_leaderboard_subset[self.metrics[1]]
        best_row = rashomon_leaderboard_subset.loc[rashomon_leaderboard_subset['weighted_sum'].idxmax()]
        best_model = best_row['model']
        return best_model
    
    def find_worst_model(self) -> str:
        '''
            Method for finding worst_model based on user passed weight values. 

            Returns:
                str : Model name that minimizes the sum : weight1*metrics[0] + weight2*metrics[1].
                      If many models have the same sum value, base model is the first one in the leaderboard.
        '''
        weight1 = self.weights.iloc[0]
        weight2 = self.weights.iloc[1]

        if not np.isclose(weight1+weight2, 1, atol=1e-6) or weight1<0 or weight2<0:
            raise ValueError('Passed weights should be >= 0 and sum to 1')
        
        rashomon_leaderboard_subset = self.leaderboard[self.leaderboard['model'].isin(self.rashomon_set)].copy()
        rashomon_leaderboard_subset['weighted_sum'] = weight1 * rashomon_leaderboard_subset[self.metrics[0]] + weight2* rashomon_leaderboard_subset[self.metrics[1]]
        worst_row = rashomon_leaderboard_subset.loc[rashomon_leaderboard_subset['weighted_sum'].idxmin()]
        worst_model = worst_row['model']
        return worst_model
    
    def find_same_score_as_base(self) -> Tuple[int, list]:
        '''
            Method for finding how many models from the leaderboard have the same value weighted sum of two passed metrics.

            Returns:
                Tuple[int, list]:
                - Number of models (excluding base_model)
                - List of model names (excluding base_model)
        '''
        weight1 = self.weights.iloc[0]
        weight2 = self.weights.iloc[1]

        if not np.isclose(weight1 + weight2, 1, atol=1e-6):
            raise ValueError("Weights must sum to 1.")

        leaderboard_subset = self.leaderboard[self.leaderboard['model'].isin(self.rashomon_set)].copy()
        leaderboard_subset['weighted_score'] = (weight1 * leaderboard_subset[self.metrics[0]] + weight2 * leaderboard_subset[self.metrics[1]])
        base_model_score = leaderboard_subset.loc[leaderboard_subset['model'] == self.base_model, 'weighted_score'].values[0]

        same_score_rows = leaderboard_subset[leaderboard_subset['weighted_score'] == base_model_score]
        same_score_rows = same_score_rows[same_score_rows['model'] != self.base_model]
        same_scores_count = same_score_rows.shape[0]
        same_scores_models = same_score_rows['model'].tolist()
        
        return same_scores_count, same_scores_models

    
    def find_weights_entropy_based(self) -> pd.Series:
        '''
            Method for finding weights based on the entropy method.

            Returns:
                pd.Series : weights with index as metric names and values as weights.
        '''

        leaderboard_subset = self.leaderboard[['model'] + self.metrics]
        leaderboard_subset = leaderboard_subset[leaderboard_subset['model'].isin(self.rashomon_set)] #only models from rashomon set (intersection)

        #step1: scaling 
        leaderboard_subset[self.metrics] = leaderboard_subset[self.metrics].apply(lambda x: x / x.sum())
        #step 2: calculating entropy for each column as -h * sum(rij *ln(rij)) where h = 1/ln(numer_of_models) and rij is a scaled cell value
        h = -1 / np.log(len(leaderboard_subset))
        leaderboard_subset[self.metrics] = leaderboard_subset[self.metrics].apply(lambda x: x * np.log(x + 1e-9)) # rij * ln(rij) 
        entropy_values = leaderboard_subset[self.metrics].sum() * h 

        # step 3: calculate weights for each column as wj = (1-ej)/sum(1-ej)
        denominator = (1-entropy_values).sum()
        weights = (1-entropy_values)/denominator
        if not np.isclose(weights.sum(), 1.0, atol=1e-6):
            raise ValueError(f'Weights do not add up to 1!: {weights}')
        
        return weights
    
    def find_base_model_entropy_based(self) -> str:
        '''
            Method for finding base_model based on entropy method.

            Returns:
                str : Model name that maximizes the sum : weight1*metrics[0] + weight2*metrics[1].
                      If many models have the same sum value, base model is the first one in the leaderboard.
        '''

        weights_series = self.find_weights_entropy_based()
        weights = weights_series.loc[self.metrics]
        best_model = self.find_base_model(weights.iloc[0], weights.iloc[1])

        return best_model
    
    def find_weights_critic_method(self) -> pd.Series:
        '''
            Method for finding weights based on the CRITIC method.
            Falls back to entropy method if CRITIC fails for any reason.

            Returns:
                str : weights with index as metric names and values as weights.
        '''

        try:
            leaderboard_subset = self.leaderboard[['model'] + self.metrics]
            leaderboard_subset = leaderboard_subset[leaderboard_subset['model'].isin(self.rashomon_set)] #only models from rashomon set (intersection)

            if len(leaderboard_subset) < 2:
                raise ValueError(f"Insufficient number of models ({len(leaderboard_subset)}) in Rashomon set. At least 2 models are required.")

            #step1: min-max normalization -> all metrics are in [0,1] range
            leaderboard_subset[self.metrics] = leaderboard_subset[self.metrics].apply(
                lambda x: (x - x.min()) / (x.max() - x.min()) if x.max() > x.min() else 0)
            
            #if all values in both columns are the same, correlation matrix cannot be calculated (there is no variation) therefore CRITIC metric cannot be applied 
            if (leaderboard_subset[self.metrics].nunique() == 1).any():
                raise ValueError(f"One or more metrics have identical values across all models: {leaderboard_subset[self.metrics].nunique()}.")
            else:
                #step 2: computing pearson correlation matrix for all metrics
                correlation_matrix = leaderboard_subset[self.metrics].corr(method='pearson')
                if correlation_matrix.isna().any().any():
                    raise ValueError("Correlation matrix contains NaN values, likely due to zero or near-zero variation in metrics.")
                
                #step 3: computing standard deviation for each metric
                standard_deviations = leaderboard_subset[self.metrics].std(ddof=0)
                if (standard_deviations == 0).any():
                    raise ValueError(f"One or more metrics have zero standard deviation: {standard_deviations}.")
                
                #step 4: calculating weights as cj = sj * sum(1 - correlation_matrix_ij) where sj is standard deviation for metric j and correlation_matrix_ij is correlation between metric j and other metrics
                c_values = standard_deviations * (1-correlation_matrix).sum(axis=1)
                weights = c_values / c_values.sum()

                if not np.isclose(weights.sum(), 1.0, atol=1e-6):
                    raise ValueError(f'Weights do not add up to 1!: {weights}')
        
            return weights
        except Exception as e:
            warnings.warn(f"CRITIC method not applicable to this task: {e}. Falling back to default: entropy method.")
            return self.find_weights_entropy_based()

    
    def find_base_model_critic_based(self) -> str:
        '''
            Method for finding base_model based on CRITIC method.

            Returns:
                str : Model name that maximizes the sum : weight1*metrics[0] + weight2*metrics[1].
                      If many models have the same sum value, base model is the first one in the leaderboard.
        '''

        weights_series = self.find_weights_critic_method()
        weights = weights_series.loc[self.metrics]
        best_model = self.find_base_model(weights.iloc[0], weights.iloc[1])

        return best_model
    

    def get_rashomon_metrics(self, delta: float = 0.1) -> Dict[str, float]:
        '''
            Method for getting all Rashomon Intersection metrics in a dictionary format.

            Args:
                delta : threshold for probabilistic ambiguity and discrepancy

            Returns:
                Dict[str, float] : Dictionary with Rashomon Intersection metrics where keys are metric names.
        '''
        capacity_values = [self.rashomon_capacity(i) for i in range(self.determine_number_of_samples())]

        metrics = {
            'Rashomon Set Size': len(self.rashomon_set),
            'Task Type': self.determine_task_type(),
            'Number of classes': self.determine_number_of_classes(),
            'Rashomon Ratio': self.rashomon_ratio(),
            'Pattern Rashomon Ratio': self.pattern_rashomon_ratio(),
            'Ambiguity': self.binary_ambiguity() if self.determine_task_type() == 'binary' else self.multiclass_ambiguity(),
            'Discrepancy': self.binary_discrepancy() if self.determine_task_type() == 'binary' else self.multiclass_discrepancy(),
            'Probabilistic Ambiguity': self.probabilistic_abiguity(delta) if self.determine_task_type() == 'binary' else None,
            'Probabilistic Discrepancy': self.probabilistic_discrepancy(delta) if self.determine_task_type() == 'binary' else None,
            'VPRs' : self.viable_prediction_range() if self.determine_task_type() =='binary' else None,
            'Agreement rates list': self.agreement_rate(),
            'Percent agreement dict': self.percent_agreement(),
            'Mean Rashomon Capacity': np.mean(capacity_values),
            'Min Rashomon Capacity': np.min(capacity_values),
            'Max Rashomon Capacity': np.max(capacity_values),
            'Std Rashomon Capacity': np.std(capacity_values)
        }
        return metrics
    
    
    def summarize_rashomon(self, delta: float = 0.1):
        '''
            Method for printing all calculated metrics for Rashomon Intersection.
        '''
        
        metrics = self.get_rashomon_metrics(delta=delta)

        print(f"\n=== Rashomon Intersection Set Summary ===\n")
        print(f"Selected metrics: {self.metrics[0]}, {self.metrics[1]}")
        print(f"Determined task type is: {metrics['Task Type']}")
        print(f"Number of classes: {metrics['Number of classes']}")
        print(f"Computed weights: {self.weights.iloc[0]}, {self.weights.iloc[1]}")
        print(f"The size of Rashomon set for given epsilon is: {metrics['Rashomon Set Size']}.")

        print(f'Base model for given metrics ({self.metrics[0]}, {self.metrics[1]}) is {self.base_model}.')
        #number of models with the same score as the base model
        print(f'There were {self.find_same_score_as_base()[0]} other models found with the same weighted sum of metrics as base model. These models are: {self.find_same_score_as_base()[1]}')

        #Rashomon Ratios
        print(f'Rashomon Ratio metric value is {self.rashomon_ratio():.2f}')
        print(f'Pattern Rashomon Ratio metric value is {self.pattern_rashomon_ratio():.2f}')

        if metrics['Task Type'] == 'binary':
            print(f"Binary Ambiguity: {metrics['Ambiguity']:.2f}")
            print(f"Binary Discrepancy: {metrics['Discrepancy']:.2f}")
            print(f"Probabilistic Ambiguity (delta={delta}): {metrics['Probabilistic Ambiguity']:.2f}")
            print(f"Probabilistic Discrepancy (delta={delta}): {metrics['Probabilistic Discrepancy']:.2f}")
            #print(f"Viable Prediction Range (VPRs): {metrics['VPRs']}")
        else:
            print(f"Multiclass Ambiguity: {metrics['Ambiguity']:.2f}")
            print(f"Multiclass Discrepancy: {metrics['Discrepancy']:.2f}")

        # Agreement rates
        print(f"\nMax agreement rate: {np.max(metrics['Agreement rates list'])}")
        print(f"\nMin agreement rate: {np.min(metrics['Agreement rates list'])}")
        print(f"\nStd agreement rate: {np.std(metrics['Agreement rates list'])}")
        print("\n")
        print("Agreement percents for all models in the Rashomon Intersection: \n")
        for key, value in metrics['Percent agreement dict'].items():
            print(f"{key} : {value:.2f}\n")

        # Rashomon Capacity
        print(f"Mean Rashomon Capacity: {metrics['Mean Rashomon Capacity']:.2f}")
        print(f"Min Rashomon Capacity: {metrics['Min Rashomon Capacity']:.2f}")
        print(f"Max Rashomon Capacity: {metrics['Max Rashomon Capacity']:.2f}")
        print(f"Std Rashomon Capacity: {metrics['Std Rashomon Capacity']:.2f}")
    


import streamlit as st
import pickle
import sys
import streamlit as st
import pickle
import sys
from pathlib import Path
import re
import streamlit.components.v1 as components
import plotly.graph_objects as go
BASE_DIR = Path(__file__).parent
CSS_PATH = BASE_DIR / "assets" / "style.css"

plot_backgroud_color = "white"
plot_font = dict(color='#989898', size=12)
plot_title_font = dict(
    color="#6B7F8A",
    family="Inter, Segoe UI, sans-serif",  
    size=16,
    style="normal",
    weight=300            
)



#Rashomon Set dashboard for binary task type

def run_app_from_file(plots_file):
    with open(plots_file, "rb") as f:
        plots, descriptions = pickle.load(f)
    
    st.set_page_config(page_title="Rashomon Dashboard", layout="wide")

    with open(CSS_PATH) as f:
        css = f.read()

    st.markdown(f"<style>{css}</style>", unsafe_allow_html=True)

    x0, x1, y0, y1 = 0, 1, 0, 1
    r = 0.02
    path = f'M{x0+r},{y0} ' \
       f'L{x1-r},{y0} Q{x1},{y0} {x1},{y0+r} ' \
       f'L{x1},{y1-r} Q{x1},{y1} {x1-r},{y1} ' \
       f'L{x0+r},{y1} Q{x0},{y1} {x0},{y1-r} ' \
       f'L{x0},{y0+r} Q{x0},{y0} {x0+r},{y0} Z'
    
    x0v, x1v, y0v, y1v = 0, 1, -0.02, 1.05
    path_violin = f'M{x0v+r},{y0v} ' \
       f'L{x1v-r},{y0v} Q{x1v},{y0v} {x1v},{y0v+r} ' \
       f'L{x1v},{y1v-r} Q{x1},{y1v} {x1v-r},{y1v} ' \
       f'L{x0v+r},{y1v} Q{x0v},{y1v} {x0v},{y1v-r} ' \
       f'L{x0v},{y0v+r} Q{x0v},{y0v} {x0v+r},{y0v} Z'

    title_container = st.container()
    title_container.markdown('<div class="dashboard-title">Rashomon Intersection Analysis</div>', unsafe_allow_html=True)
    st.markdown("<br>", unsafe_allow_html=True)
    
    #TABLES ------------------------------------------------------------------------------------------------------
    with st.container():
        rashomon_set_col, pareto_col = st.columns(2)
        
        with rashomon_set_col:
            
            rashomon_table = plots["generate_rashomon_set_table"]
            rashomon_table.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    ),
                    ],
                xaxis=dict(showgrid=False, zeroline=False),
                yaxis=dict(showgrid=False, zeroline=False),
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=450,
                height=350
            )
            rashomon_table.update_layout(

                margin=dict(t=10, l=10, r=10, b=0)
            )
            st.markdown('<div class = "section_title"> Models present in the Rashomon Intersection</div>', unsafe_allow_html=True)
            st.markdown(f'<div class="section_description">This section presents all models icluded in the Rashomon Intersection. Firstly, two separate Rashomon Sets for base metrics : {descriptions["base_metric_return"][0]} and {descriptions["base_metric_return"][1]} were obtained. Only models present in both sets were included in the Rashomon Intersection illustrated by the table below.  </div>', unsafe_allow_html=True)

            with st.container(key = "basics"):
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                            <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px; ">Base model : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["base_model_return"]}</span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                             <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px;">Base metrics : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["base_metric_return"][0]},  {descriptions["base_metric_return"][1]} </span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
                st.markdown(
                        f"""
                        <div style="font-size:15px; margin-bottom:10px;">
                            <span style="color: #426c85;font-family: 'Inter', sans-serif; font-weight: 500; letter-spacing: 0.2px;">Weighted sum method : </span>
                            <span style="color:#989898; font-family: 'Inter', sans-serif;"> {descriptions["weighted_method_return"]}</span>
                        </div>
                        """,
                        unsafe_allow_html=True
                )
            with st.container(key = "table_container", height="stretch"):
                st.plotly_chart(rashomon_table, width='stretch')
                st.markdown(f'<div class="plots_descr rashomon-table-descr">{descriptions["generate_rashomon_set_table"]}</div>', unsafe_allow_html=True)
        
        
        with pareto_col:
            pareto_plot, pareto_descr = plots['pareto_front_plot'], descriptions['pareto_front_plot']

            pareto_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=450,
                height=450,
                font=plot_font,
                legend=dict(
                font=plot_font
            )
            )
            pareto_plot.update_layout(
                title={
                    'text': "Pareto Front",
                    'y':0.97,
                    'x':0.05,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=40, l=3, r=3, b=25)
            )

            st.markdown('<div class = "section_title">Pareto Front with all models</div>', unsafe_allow_html=True)
            st.markdown('<div class = "section_description"> The plot below presents all models from the leaderboard with highlighted models included in the Pareto front obtained for both metrics. Note that models in the Pareto Front do not have to match models present in the Rashomon Intersection, as the selection procedure for the Intersection was different.</div>', unsafe_allow_html=True)
            with st.container(key = "pareto", height="stretch"):
                st.plotly_chart(pareto_plot, width='stretch')
                st.markdown(f'<div class="plots_descr">{pareto_descr}</div>', unsafe_allow_html=True)

    with st.container():
        st.markdown('<div class = "section_title">Rashomon Intersection analysis</div>', unsafe_allow_html=True)
        st.markdown('<div class = "section_description">In the following sections the analysis is focused on the Rashomon Intersection, not the models in the Pareto Front.</div>', unsafe_allow_html=True)

    with st.container():
        gauge_col, venn_col = st.columns([40,60])

        with gauge_col:
            st.markdown('<div class = "section_title">Rashomon Intersection size</div>', unsafe_allow_html=True)
            gauge_size, gauge_size_descr = plots["set_size_indicator"], descriptions["set_size_indicator"]

            gauge_size.update_layout(
                height=270,
                width=350,
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                margin=dict(t=50, l=80, r=80, b=10),
                font=plot_font,

            )
            gauge_rashomons, gauge_rashomons_descr = plots["generate_gauge_for_metrics"], descriptions["generate_gauge_for_metrics"]

            gauge_rashomons.update_layout(
                height=220,
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                font=plot_font,
            )

            with st.container(key = "gauge"):
                st.markdown(f'<div class = "section_description">The following chart illustrates the fraction of models from the leaderboard that are included in the Rashomon Intersection.</div>', unsafe_allow_html=True)
                st.plotly_chart(gauge_size)
                st.markdown(f'<div class = "section_description">Additionally, the numbers of models from separate Rashomon Sets included in the Intesection are represented by the plots below.</div>', unsafe_allow_html=True)
                st.plotly_chart(gauge_rashomons)

        
        with venn_col:
            st.markdown('<div class = "section_title">Venn diagram </div>', unsafe_allow_html=True)
            venn_plot, venn_descr = plots['plot_venn_diagram'], descriptions['plot_venn_diagram']
            venn_plot.update_layout(
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                font=plot_font,
                height = 400
            )

            with st.container(key = "venn", height = 'stretch'):
                st.markdown(f'<div class = "venn_description">{venn_descr}</div>', unsafe_allow_html=True)
                st.plotly_chart(venn_plot)

    # -------------------------------------- SAME AS RASHOMON SET -----------------------------------------------------------------

    with st.container():
        st.markdown(
            '<div class="section_title">Rashomon Ratio and Pattern Ratio metrics</div>',
            unsafe_allow_html=True
        )
        rashomon_ratio_col,pattern_ratio_col = st.columns(2)

        with pattern_ratio_col:              
            # # PATTERN RATIO INDICATOR
            pattern_indicator, pattern_indicator_descr = plots['pattern_ratio_indicator'], descriptions['pattern_ratio_indicator']
            pattern_indicator.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=75,
                font=plot_font,
            )
            pattern_indicator.update_layout(
                title={
                    'text': "Pattern Rashomon Ratio",
                    'y':0.97,
                    'x':0.05,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=10, r=10, b=25)
            )
    
            with st.container(key = "pattern_white_container", height="stretch"):
                    st.plotly_chart(pattern_indicator, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{pattern_indicator_descr}</div>',
                        unsafe_allow_html=True
                    )
    

        with rashomon_ratio_col:
        # RASHOMON RATIO 
            ratio_indicatior, ratio_indicator_descr = plots['rashomon_ratio_indicator'], descriptions['rashomon_ratio_indicator']
            ratio_indicatior.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                font=plot_font,
            )
            ratio_indicatior.update_layout(
                title={
                    'text': "Rashomon Ratio",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                width=75,
                height=75,
                margin=dict(t=30, l=10, r=10, b=25)
            )

            with st.container(key = "ratio_white_container", height="stretch"):
                    st.plotly_chart(ratio_indicatior, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{ratio_indicator_descr}</div>',
                        unsafe_allow_html=True
                    )
           
        
        

# RATIOS VS EPSILONS
    with st.container():
        col1, col2 = st.columns(2)
        with col1:
             # # RASHOMON RATIO VS EPSILON
            ratio_vs_eps_plot, ratio_vs_eps_descr = plots['rashomon_ratio_vs_epsilon'], descriptions['rashomon_ratio_vs_epsilon']
            ratio_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor= plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=300,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Rashomon Ratio vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=50, l=10, r=10, b=0)
            )
            


            with st.container(key = "ratio_epsilon_white_container", height="stretch"):
                    st.plotly_chart(ratio_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{ratio_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )
        with col2:
            # PATTERN VS EPSILON
            pattern_vs_eps_plot, pattern_vs_eps_descr = plots['pattern_rashomon_ratio_vs_epsilon'], descriptions['pattern_rashomon_ratio_vs_epsilon']
            pattern_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                width=75,
                height=300,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Pattern Ratio vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=50, l=10, r=10, b=0)
            )

            with st.container(key = "pattern_epsilon_white_container", height="stretch"):
                    st.plotly_chart(pattern_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{pattern_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )


    #AMBIGUITY AND DISCREPANCY -------------------------------------------------------------------------------------------------------
    #with st.container(key="amb_discr"):

    with st.container():
        st.markdown(
            '<div class="section_title">Ambiguity and Discrepancy metrics</div>',
            unsafe_allow_html=True
        )
        descr= """Ambiguity can be interpreted as the proportion of observations that would receive a different prediction if we were to choose another model from the Rashomon Set, while
       Discrepancy is the maximum proportion of individuals whose predictions would change if we choose a different competing model from the Rashomon Set."""
        st.markdown(
            f'<div class="section_description">{descr}</div>',
            unsafe_allow_html=True
        )
        st.markdown(
            """
            <style>
            [data-baseweb="checkbox"] [data-testid="stWidgetLabel"] p {
                /* Styles for the label text */
                color: #8B9DA8;
                font-family: 'Inter', sans-serif; 
                font-size: 16px; 
                font-weight: 300 !important;
                letter-spacing: 0.1px;
                line-height: 1.5; 

            }
            </style>
            """,
            unsafe_allow_html=True

        )



        lolipop_col, amb_col, discr_col = st.columns([1.1,1.5,1.5])
        with st.container():
            with lolipop_col:
                
                
                
                amb_discr_plot, amb_discr_descr = plots['lolipop_ambiguity_discrepancy'], descriptions['lolipop_ambiguity_discrepancy']

                amb_discr_plot.update_layout(
                    paper_bgcolor= plot_backgroud_color,
                    plot_bgcolor=plot_backgroud_color,
                    height=250,
                    font=plot_font,
                    xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    legend=dict(font=dict(color="#989898", size=10)),
                    title={
                    'text': "Ambiguity and Discrepancy values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                    },
                )
                with st.container(key = "lolipop_container"):
                    st.plotly_chart(amb_discr_plot)
                    st.markdown(
                        f'<div class ="plots_descr">{amb_discr_descr}</div>',
                        unsafe_allow_html=True
                    )

            with amb_col:
                amb_vs_eps_plot, amb_vs_eps_descr = plots['ambiguity_vs_epsilon'], descriptions['ambiguity_vs_epsilon']
                amb_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=250,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Ambiguity vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=5, r=5, b=5)
                )

                with st.container(key="ambig_epsilon_container", height="stretch"):
                    st.plotly_chart(amb_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{amb_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )

            with discr_col:
                
                discr_vs_eps_plot, discr_vs_eps_descr = plots['discrepancy_vs_epsilon'], descriptions['discrepancy_vs_epsilon']
                discr_vs_eps_plot.update_layout(
                shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=250,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Discrepancy vs Epsilon values",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=30, l=5, r=5, b=5)
                )

                with st.container(key = "discr_epsilon_container", height="stretch"):
                    st.plotly_chart(discr_vs_eps_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{discr_vs_eps_descr}</div>',
                        unsafe_allow_html=True
                    )

    #RASHOMON CAPACITY AND PER SAMPLE -------------------------------------------------------------------------------------------------------
    with st.container():
        left, right = st.columns([45,55], vertical_alignment="top")
        with right:
            st.markdown('<div class = "section_title">Single sample analysis</div>', unsafe_allow_html=True)
            descr= f"""
            This section presents a detailed analysis of a randomly selected sample from the dataset.
            <br>
            If you wish to explore a specific sample, try experimenting with the methods (rashomon_capacity_for_sample and proba_probabilities_for_sample) available in the Visualizer class instance.
            </br>
            """
            st.markdown(
                    f'<div class ="section_description">{descr}</div>',
                    unsafe_allow_html=True
                )
            
        with left:
            st.markdown(
                '<div class="section_title">Rashomon Capacity metric</div>',
                unsafe_allow_html=True
            )
            no_classes = descriptions["number_of_classes_return"]
            descr= f"""Rashomon Capacity is a predictive multiplicity metric that measures the diversity of model predictions for each sample.
            <br> 
            <strong>Values are between [1, {no_classes}]</strong>, where 1 indicates do predictive multiplicty present, and {no_classes} represents the highest possible diversity among model predictions.
            </br>
            """
            st.markdown(
                f'<div class="section_description">{descr}</div>',
                unsafe_allow_html=True
            )
    with st.container():
        
        rashomon_capacity_col, per_sample_col = st.columns([45,55], vertical_alignment="top")

        with rashomon_capacity_col:
            with st.container():
                capacity_hist, capacity_hist_descr = plots['rashomon_capacity_distribution'], descriptions['rashomon_capacity_distribution']
                capacity_hist.update_layout(
                    shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                ],
                paper_bgcolor=plot_backgroud_color,
                plot_bgcolor=plot_backgroud_color,
                height=230,
                font=plot_font,
                xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                legend=dict(font=dict(color="#989898", size=10)),
                title={
                    'text': "Rashomon Capacity distribution",
                    'y':0.97,
                    'x':0.02,
                    'xanchor': 'left',
                    'yanchor': 'top',
                    'font': plot_title_font
                },
                margin=dict(t=40, l=5, r=10, b=5)
                )

                capacity_boxplot, capacity_boxplot_descr = plots['rashomon_capacity_distribution_by_class'], descriptions['rashomon_capacity_distribution_by_class']
                capacity_boxplot.update_layout(
                    shapes=[
                    dict(
                        type="path",
                        path=path,
                        xref="paper", yref="paper",
                        line=dict(color="#6C929E", width=1),
                        fillcolor="rgba(0,0,0,0)"
                    )
                    ],
                    paper_bgcolor = plot_backgroud_color,
                    plot_bgcolor=plot_backgroud_color,
                    height=230,
                    font=plot_font,
                    xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
                    legend=dict(font=dict(color="#989898", size=10)),
                    title={
                        'text': "Rashomon Capacity distribution by class",
                        'y':0.97,
                        'x':0.02,
                        'xanchor': 'left',
                        'yanchor': 'top',
                        'font': plot_title_font
                    },
                    margin=dict(t=40, l=5, r=10, b=5)
                    )

                #histogram container for plot and description
              
                with st.container(key='capacity-hist'):
                    st.plotly_chart(capacity_hist, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_hist_descr}</div>',
                        unsafe_allow_html=True
                    )

                #boxplot container for plot and description
                with st.container(key='capacity-boxplot', height="stretch"):
                    st.plotly_chart(capacity_boxplot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_boxplot_descr}</div>',
                        unsafe_allow_html=True
                    )
                    

        with per_sample_col:
                
            proba_probabilities_plot, proba_probabilities_descr = plots['proba_probabilities_for_sample'], descriptions['proba_probabilities_for_sample']

            proba_probabilities_plot.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=400,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=10)),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=10)),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Predicted probabilities for sample across all models in the Rashomon Set",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=30, l=5, r=5, b=5)
            )

            capacity_per_sample, capacity_per_sample_descr = plots['rashomon_capacity_for_sample'], descriptions['rashomon_capacity_for_sample']

            capacity_per_sample.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=100,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Rashomon Capacity value for sample",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=60, l=5, r=5, b=10)
            )

            
           
            #proba probabilities container for plot and description
            with st.container(key='per_sample', height="stretch"):
                st.plotly_chart(proba_probabilities_plot, width='stretch')
                st.markdown(
                    f'<div class="plots_descr">{proba_probabilities_descr}</div>',
                    unsafe_allow_html=True
                )

            #rashomon capacity for sample container for plot and description
                with st.container(key="capacity_per_sample_padding"):
                    st.plotly_chart(capacity_per_sample, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{capacity_per_sample_descr}</div>',
                        unsafe_allow_html=True
                    )


    #PERCENT AGREEMENT AND FEATURE IMPORTANCE -------------------------------------------------------------------------------------------------------

    with st.container():
        
        percent_agreement_bar_plot, percent_agreement_bar_descr = plots['percent_agreement_barplot'], descriptions['percent_agreement_barplot']
        percent_agreement_bar_plot.update_layout(
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=350,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Percent Agreement Rates distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=10, b=5)
            )
        
        cohens_kappa_barplot, cohens_kappa_barplot_descr = plots['cohens_kappa_diverging_barplot'], descriptions['cohens_kappa_diverging_barplot']
        cohens_kappa_barplot.update_layout(
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=350,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Cohen's Kappa distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=10, b=5)
            )
        
        agreement_rates_violin_plot, agreement_rates_violin_descr = plots['agreement_rates_density'], descriptions['agreement_rates_density']
        agreement_rates_violin_plot.update_yaxes(range=[-0.005,1.005], tickvals=[0, 0.2, 0.4, 0.6, 0.8, 1], showgrid=True)
        agreement_rates_violin_plot.update_layout(
            shapes=[
            dict(
                type="path",
                path=path_violin,
                xref="paper", yref="paper",
                line=dict(color="#6C929E", width=1),
                fillcolor="rgba(0,0,0,0)"
            )
            ],
            paper_bgcolor=plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=390,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898")),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898"), showgrid=True),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Agreement Rates distribution",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=5, r=10, b=5)
            )
        
        fi_table, fi_table_descr = plots['feature_importance_table'], descriptions['feature_importance_table']
        fi_heatmap, fi_heatmap_descr = plots['feature_importance_heatmap'], descriptions['feature_importance_heatmap']

        fi_table.update_traces(columnwidth=[1,2,2])
        fi_table.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                ),
                ],
            xaxis=dict(showgrid=False, zeroline=False),
            yaxis=dict(showgrid=False, zeroline=False),
            paper_bgcolor= plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=139
        )
        fi_table.update_layout(
            title={
                'text': "Top 3 Feature Importance in Rashomon Set",
                'y':0.97,
                'x':0.05,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=40, l=10, r=10, b=10)
        )

        heatmap_is_empty = (len(fi_heatmap.data) == 0 or fi_heatmap.data[0].z is None or len(fi_heatmap.data[0].z) == 0)
        if not heatmap_is_empty:
            fillcolor = 'rgba(108,146,158,0.2)'
        else:
            fillcolor = 'rgba(0,0,0,0)'
                
        fi_heatmap.update_layout(
            shapes=[
                dict(
                    type="path",
                    path=path,
                    xref="paper", yref="paper",
                    line=dict(color="#6C929E", width=1),
                    fillcolor="rgba(0,0,0,0)"
                ),
                dict(
                    type="rect",
                    xref="paper", yref="paper",
                    x0=0, y0=0, x1=1, y1=1,
                    fillcolor=fillcolor,
                    line=dict(width=0),
                    layer="below"   #background below the heatmap to show borders between cells
                )
            ],
            paper_bgcolor= plot_backgroud_color,
            plot_bgcolor=plot_backgroud_color,
            height=570,
            font=plot_font,
            xaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=8)),
            yaxis=dict(title_font=dict(color="#989898"), tickfont=dict(color="#989898", size=8)),
            legend=dict(font=dict(color="#989898", size=10)),
            title={
                'text': "Feature Importance Rank",
                'y':0.97,
                'x':0.02,
                'xanchor': 'left',
                'yanchor': 'top',
                'font': plot_title_font
            },
            margin=dict(t=50, l=0, r=5, b=5),
            )

        percent_agreement_col, feature_importance_col = st.columns([45, 55], vertical_alignment='top')
        agreement_rates_col, fi_heatmap_col = st.columns([45, 55], vertical_alignment='bottom')

        with st.container():

            with percent_agreement_col:
                st.markdown('<div class = "section_title">Models agreement metrics</div>', unsafe_allow_html=True)
                descr = '''
                    This section shows how models in the Rashomon Set agree with the base model using Percent Agreement or Cohen’s Kappa metrics.
                    '''
                st.markdown(f'<div class="section_description no_padding">{descr}</div>', unsafe_allow_html=True)

                metric_option = st.radio(
                    "Select metric:",
                    ["Percent Agreement", "Cohen's Kappa"],
                    horizontal=True,
                    label_visibility="collapsed"
                )

                if metric_option=="Percent Agreement":
                    st.markdown('<div class = "section_title">Percent Agreement</div>', unsafe_allow_html=True)
                    descr = '''
                        Percent Agreement is defined as the proportion of observations for which each model from the Rashomon Set predicted the same class as the base model.
                        '''
                    st.markdown(f'<div class="section_description less_padding">{descr}</div>', unsafe_allow_html=True)
                    with st.container(key = "percent_agreement"):
                        st.plotly_chart(percent_agreement_bar_plot, width='stretch')
                        st.markdown(f'<div class="plots_descr">{percent_agreement_bar_descr}</div>', unsafe_allow_html=True)

                elif metric_option=="Cohen's Kappa":
                    st.markdown('<div class = "section_title">Cohens Kappa</div>', unsafe_allow_html=True)
                    descr = '''
                        Cohen's Kappa measures how much each model agrees with the base model beyond chance.
                        Values range from -1 (complete disagreement) to 1 (perfect agreement), with 0 meaning chance-level agreement.
                        '''
                    st.markdown(f'<div class="section_description less_padding">{descr}</div>', unsafe_allow_html=True)
                    with st.container(key = "cohens_kappa_barplot"):
                        st.plotly_chart(cohens_kappa_barplot, width='stretch')
                        st.markdown(f'<div class="plots_descr">{cohens_kappa_barplot_descr}</div>', unsafe_allow_html=True)

            with feature_importance_col:
                st.markdown('<div class = "section_title">Feature Importance</div>', unsafe_allow_html=True)
                descr = '''
                This section provides an overview of the most important features identified across the models included in the Rashomon Set, highlighting which features contribute most to the predictions of each model.
                <br>
                <em>Note that this section presents information for models with feature importnace data available. 
                If feature importance was not selected or was unavailable for certain models, no information is displayed.</em><br>
                '''
                st.markdown(f'<div class="section_description less_padding">{descr}</div>', unsafe_allow_html=True)
                with st.container(key='fi_table'):
                    st.plotly_chart(fi_table, width='stretch')
                    fi_table_description_update = fi_table_descr + " <br>If the base model does not provide feature importance values, a ‘–’ will be displayed to indicate that this information is unavailable."
                    st.markdown(f'<div class = "plots_descr">{fi_table_description_update}</div>', unsafe_allow_html=True)
                st.markdown(f'<div class="section_description heatmap-descr">{fi_heatmap_descr}</div>', unsafe_allow_html=True)

        with st.container():
            with agreement_rates_col:
                st.markdown('<div class ="section_title">Agreement Rates</div>', unsafe_allow_html=True)
                descr = '''
                For each observation, the Agreement Rate quantifies the proportion of models that produce the same class prediction as the base model.
                '''
                st.markdown(
                        f'<div class="section_description">{descr}</div>',
                        unsafe_allow_html=True
                    )

                #violin container for plot and description
                with st.container(key='agreement_rates_violin'):
                    st.plotly_chart(agreement_rates_violin_plot, width='stretch')
                    st.markdown(
                        f'<div class="plots_descr">{agreement_rates_violin_descr}</div>',
                        unsafe_allow_html=True
                    )
            with fi_heatmap_col:
                #feature importance heatmap
                with st.container(key = 'fi_heatmap', height="stretch"):
                    st.plotly_chart(fi_heatmap, width='stretch')


if __name__ == "__main__":
    plots_file = sys.argv[1]   
    run_app_from_file(plots_file)


import pandas as pd
import time
import os
import sys
import inspect 
import random
import subprocess
import psutil
from pathlib import Path
import importlib.resources as pkg_resources
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
from rashomon_analysis.rashomon_set import *
from rashomon_analysis.rashomon_intersection import *
from rashomon_analysis.visualizers.rashomon_visualizer import *
from rashomon_analysis.visualizers.intersection_visualizer import *
from app.rashomon_binary_app import *
from rashomon_analysis.pipelines.builder_abstract import *
import pickle
import warnings
import logging
BLUE_BOLD = "\033[1;34m"
RESET = "\033[0m"

logging.basicConfig(
    level=logging.INFO,
    format=f"{BLUE_BOLD}%(message)s{RESET}"
)
PROJECT_ROOT = Path(__file__).resolve().parents[3] #pipelines -> rahsomon_analysis ->src -> root
# RASHOMON_BINARY_APP_PATH = PROJECT_ROOT/"src"/"app"/"rashomon_binary_app.py"
# RASHOMON_MULTICLASS_APP_PATH = PROJECT_ROOT/"src"/"app"/"rashomon_multiclass_app.py"
# INTERSECTION_BINARY_APP_PATH = PROJECT_ROOT/"src"/"app"/"intersection_binary_app.py"
# INTERSECTION_MULTICLASS_APP_PATH = PROJECT_ROOT/"src"/"app"/"intersection_multiclass_app.py"

INTERSECTION_BINARY_APP_PATH = Path(pkg_resources.files("app") / "intersection_binary_app.py")
INTERSECTION_MULTICLASS_APP_PATH = Path(pkg_resources.files("app") / "intersection_multiclass_app.py")
RASHOMON_BINARY_APP_PATH = Path(pkg_resources.files("app") / "rashomon_binary_app.py")
RASHOMON_MULTICLASS_APP_PATH = Path(pkg_resources.files("app") / "rashomon_multiclass_app.py")

PIPELINES_DIR = Path(__file__).resolve().parent
TMP_DIR = PIPELINES_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)  

class BuildRashomonFromConverted(Builder):

    def __init__(self, converter_results_path: Path, base_metric : str, delta : float = 0.1):
        '''
            Initializes the Pipeline from converted results in order to create the Rashomon Set and it's visualizations.
            Args:
            converter_results_path : path to a folder where all objects returned by Converter.convert() are stored. 
            Should contain files such as : leaderboard.csv, predictions_dict.pkl, proba_predictions_dict.pkl, y_true.csv and feature_importance_dict.pkl (optional)

            base_metric : evaluation metric to be used while creating the Rashomon Set

            delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used

        '''

        logging.info("Initializing Rashomon Builder...")
        self.METRICS = ['accuracy', 'balanced_accuracy', 'f1', 'f1_macro', 'f1_micro', 'f1_weighted',
            'roc_auc', 'roc_auc_ovo', 'roc_auc_ovo_macro', 'roc_auc_ovo_weighted', 'roc_auc_ovr', 
            'roc_auc_ovr_macro', 'roc_auc_ovr_micro', 'roc_auc_ovr_weighted', 'average_precision', 
            'precision', 'precision_macro', 'precision_micro', 'precision_weighted', 'recall', 
            'recall_macro', 'recall_micro', 'recall_weighted']
        
        logging.info("Arguments validation...")
        if not os.path.isdir(converter_results_path):
            raise ValueError(f"Folder path : {converter_results_path} does not exist. Provide a valid path")
        
        mandatory_files = ["leaderboard.csv", "predictions_dict.pkl", "proba_predictions_dict.pkl", "y_true.csv"]
        if not all((converter_results_path/file).is_file() for file in mandatory_files):
            raise ValueError(f"There re missing files in the provided directory. Make sure all files are present : {mandatory_files}")
        
        leaderboard = pd.read_csv( converter_results_path / "leaderboard.csv")
        y_true = pd.read_csv(converter_results_path/ "y_true.csv")
        with open(converter_results_path/ "predictions_dict.pkl", "rb") as f:
            predictions_dict = pickle.load(f)
        with open(converter_results_path / "proba_predictions_dict.pkl", "rb") as f:
            proba_predictions_dict = pickle.load(f)

        if (converter_results_path/"feature_importance_dict.pkl").is_file():
            with open(converter_results_path / "feature_importance_dict.pkl", "rb") as f:
                feature_importance_dict = pickle.load(f)
        else:
            feature_importance_dict = None


        logging.info("Arguments validated, assigning attributes...")
        self.leaderboard = leaderboard
        self.predictions = predictions_dict
        self.proba_predictions = proba_predictions_dict
        self.feature_importances = feature_importance_dict
        self.y_true = y_true
        self.base_metric = base_metric
        self.delta = delta
        self.epsilon=None

        logging.info("Rashomon Builder Initialized.")


    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Set sizes to help choose the right epsilon argument.
        '''
        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Set size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()

    def visualize_rashomon_set_volume(self):
        '''
        Method for visualising Rashomon set size depending on epsilon
        '''
        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, 10.0)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_set.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, epsilon)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()
    
    def set_epsilon(self, epsilon:float):
        '''
            Method for setting the value of Epsilon parameter to be used while creating the Rashomon Set object. 
            Is a necessary step before the build() method. 
        '''
        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")
    
    def build(self, launch_dashboard : bool = True):
        '''
            Primary method for creating  the Rashomon Set, Visualier and launching a final dashboard.
            Parameters: 
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard automatically after building the Rashomon Set. Default = True
            Returns :
                RashomonSet : created Rashomon Set object
                Visualizer : Visualizer object created based on the Rashomon Set. 
        '''
        logging.info("Building Pipeline...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        
        logging.info('Creating Rashomon Set and Visualizer...')
        rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, self.epsilon)
        visualizer = Visualizer(rashomon_set, self.y_true)

        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_set.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr

            elif rashomon_set.task_type =="multiclass":
                method_names = visualizer.multiclass_methods
            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                    func = getattr(visualizer, method)
                    sig = inspect.signature(func)
                    params = sig.parameters

                    #if the method needs parameters (random sample index - if new methods have other parameters need to provide them here !):
                    if len(params)>0:
                        if "sample_index" in params:
                            plot, descr = func(sample_index=random_idx)
                        else: 
                            raise ValueError(f"Method {method} needs unsupported parameters")
                    else: plot, descr = func()
                    plots[method] = plot
                    descriptions[method] = descr
            logging.info("Rashomon Set and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info('Closing all previous Streamlit processes...')
            self.dashboard_close() # close all processes before starting a new one
            logging.info('Launching Streamlit dashboard...')
            if rashomon_set.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_set.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info("You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        
        else:
            logging.info("Rashomon Set and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_set, visualizer

    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue



        
#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------

class BuildIntersectionFromConverted(Builder):

    def __init__(self, converter_results_path: Path, metrics : list,  custom_weights : list =None, weighted_sum_method : str = 'entropy', delta : float = 0.1):
        '''
            Initializes the Pipeline from converted results to the creation of the Rashomon Set and it's visualizations.
            Args:
            converter_results_path : path to a folder where all objects returned by Converter.convert() are stored. 
            Should contain files such as : leaderboard.csv, predictions_dict.pkl, proba_predictions_dict.pkl, y_true.csv and feature_importance_dict.pkl (optional)

            metrics : evaluation metrics to be used while creating the Rashomon Intersection

            custom_weights : list of weights for custom_weights weighted sum method. Default = None 

            weighted_sum_method: weighted sum method to be used while creating the Intersection. Default = entropy

            delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used.
        '''
        logging.info("Initializing Rashomon Intersection Builder...")
        self.METRICS = ['accuracy', 'balanced_accuracy', 'f1', 'f1_macro', 'f1_micro', 'f1_weighted',
            'roc_auc', 'roc_auc_ovo', 'roc_auc_ovo_macro', 'roc_auc_ovo_weighted', 'roc_auc_ovr', 
            'roc_auc_ovr_macro', 'roc_auc_ovr_micro', 'roc_auc_ovr_weighted', 'average_precision', 
            'precision', 'precision_macro', 'precision_micro', 'precision_weighted', 'recall', 
            'recall_macro', 'recall_micro', 'recall_weighted']

        logging.info("Arguments validation...")
        if not os.path.isdir(converter_results_path):
            raise ValueError(f"Folder path : {converter_results_path} does not exist. Provide a valid path")
        
        mandatory_files = ["leaderboard.csv", "predictions_dict.pkl", "proba_predictions_dict.pkl", "y_true.csv"]
        if not all((converter_results_path/file).is_file() for file in mandatory_files):
            raise ValueError(f"There re missing files in the provided directory. Make sure all files are present : {mandatory_files}")
        
        leaderboard = pd.read_csv( converter_results_path / "leaderboard.csv")
        y_true = pd.read_csv(converter_results_path/ "y_true.csv")
        with open(converter_results_path/ "predictions_dict.pkl", "rb") as f:
            predictions_dict = pickle.load(f)
        with open(converter_results_path / "proba_predictions_dict.pkl", "rb") as f:
            proba_predictions_dict = pickle.load(f)

        if (converter_results_path/"feature_importance_dict.pkl").is_file():
            with open(converter_results_path / "feature_importance_dict.pkl", "rb") as f:
                feature_importance_dict = pickle.load(f)
        else:
            feature_importance_dict = None

        if not delta:
            raise ValueError("Delta parameter cannot be None")
        if not isinstance(delta, float):
            raise TypeError("Delta parameter must be float")
        if delta > 1 or delta <0:
            raise ValueError("Delta parameter must be from [0-1]")
        
        logging.info("Assigning attributes...")

        #if all assertions are correct 
        self.leaderboard = leaderboard
        self.predictions = predictions_dict
        self.proba_predictions = proba_predictions_dict
        self.feature_importances = feature_importance_dict
        self.y_true = y_true
        self.metrics = metrics
        self.custom_weights = custom_weights
        self.weighted_sum_method = weighted_sum_method
        self.delta = delta
        self.epsilon=None
        logging.info("Rashomon Intersection Builder initialized.")


    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Intersection sizes to help choose the right epsilon argument.
        '''
        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Intersection size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()

    def visualize_rashomon_set_volume(self):
        '''
            Method for visualising Rashomon Intersection size depending on epsilon
        '''
        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_int = RashomonIntersection(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.metrics, 10.0, self.custom_weights, self.weighted_sum_method)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_int.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonIntersection(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.metrics, epsilon, self.custom_weights, self.weighted_sum_method)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()
    
    def set_epsilon(self, epsilon:float):
        '''
            Method for setting the value of Epsilon parameter to be used while creating the Rashomon Intersection object. 
            Is a necessary step before the build() method. 
        '''

        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")
    
    def build(self, launch_dashboard : bool = True):
        '''
            Primary method for creating  the Rashomon Intersection, IntersectionVisualier and launching a final dashboard.
            Parameters: 
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard automatically after building the Rashomon Intersection. Default = True
            Returns :
                RashomonIntersection : created Rashomon Intersection object
                IntersectionVisualizer : IntersectionVisualizer object created based on the Rashomon Intersection. 
        '''
        logging.info("Building Pipeline...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        logging.info('Creating Rashomon Intersection and Visualizer...')
        rashomon_int = RashomonIntersection(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.metrics, self.epsilon, self.custom_weights, self.weighted_sum_method)
        visualizer = IntersectionVisualizer(rashomon_int, self.y_true)
        
        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_int.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr


            elif rashomon_int.task_type =="multiclass":
                method_names = visualizer.multiclass_methods

            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                    func = getattr(visualizer, method)
                    sig = inspect.signature(func)
                    params = sig.parameters

                    #if the method needs parameters (random sample index):
                    if len(params)>0:
                        # random index 
                        if "sample_index" in params:
                            plot, descr = func(sample_index=random_idx)
                        else: 
                            raise ValueError(f"Method {method} needs unsupported parameters")
                    else: plot, descr = func()
                    plots[method] = plot
                    descriptions[method] = descr
            logging.info("Rashomon Intersection and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots2.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info("Closing all previous Streamlit processes...")
            self.dashboard_close()
            logging.info("Launching Streamlit dashboard...")
            if rashomon_int.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_int.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info(f"You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        else: 
            logging.info("Rashomon Intersection and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_int, visualizer

    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

